import std/options, std/tables
import bearssl

import ../../jwt/jwa
import ../../jwt/jwk

import ../encoding

import ./hash

type
  RsaPublicKeyObj = object
    n*, e*: string
    bearKey: RsaPublicKey

  RsaPrivateKeyObj = object
    p*, q*, dp*, dq*, iq*: string
    bearKey*: RsaPrivateKey

func initRsaPublicKeyObj(n, e: string): RsaPublicKeyObj =
  result = RsaPublicKeyObj(n: n, e: e)
  result.bearKey.n = cast[ptr cuchar](result.n.cstring)
  result.bearKey.nlen = result.n.len
  result.bearKey.e = cast[ptr cuchar](result.e.cstring)
  result.bearKey.elen = result.e.len

func initRsaPrivateKeyObj(nBitLen: int, p, q, dp, dq, iq: string): RsaPrivateKeyObj =
  result = RsaPrivateKeyObj(p: p, q: q, dp: dp, dq: dq, iq: iq)
  result.bearKey.nBitLen = cast[uint32](nBitLen)
  result.bearKey.p = cast[ptr cuchar](result.p.cstring)
  result.bearKey.plen = result.p.len
  result.bearKey.q = cast[ptr cuchar](result.q.cstring)
  result.bearKey.qlen = result.q.len
  result.bearKey.dp = cast[ptr cuchar](result.dp.cstring)
  result.bearKey.dplen = result.dp.len
  result.bearKey.dq = cast[ptr cuchar](result.dq.cstring)
  result.bearKey.dqlen = result.dq.len
  result.bearKey.iq = cast[ptr cuchar](result.iq.cstring)
  result.bearKey.iqlen = result.iq.len

proc toRsaPublicKey(jwk: JWK): RsaPublicKeyObj =
  ## Convert an RSA public key in JWK format to the wrapper for BearSSL's
  ## RsaPublicKey struct.
  return initRsaPublicKeyObj(
    n = b64UrlDecode(jwk.rsaPub.n),
    e = b64UrlDecode(jwk.rsaPub.e))

proc toRsaPrivateKey(jwk: JWK): RsaPrivateKeyObj =
  ## Convert an RSA private key in JWK format to the wrapper for BearSSL's
  ## RsaPrivateKey struct.

  # TODO: JWS spec only requires a private key to have n, e, and d, as the
  # remainder can be computed form these (p, q, dp, dq, and qi). BearSSL
  # requires p, q, dp, dq, and qi (it calls iq). Because of this, we currently
  # require all values to be present in JWKs for RSA privat keys. We should add
  # the logic to compute the missing values to fully support the JWS spec.
  #
  # We also do not currently support keys with more than two prime factors.

  let sk = jwk.rsaPrv

  if sk.p.isNone or sk.q.isNone or sk.dp.isNone or sk.dq.isNone or sk.qi.isNone:
    raise newException(ValueError,
      "RSA private key must have values for: n, e, d, p, q, dp, dq, and qi")

  let n = b64UrlDecode(sk.n)
  return initRsaPrivateKeyObj(
    nBitLen = n.len * 8,
    p = b64UrlDecode(sk.p.get),
    q = b64UrlDecode(sk.q.get),
    dp = b64UrlDecode(sk.dp.get),
    dq = b64UrlDecode(sk.dq.get),
    iq = b64UrlDecode(sk.qi.get))

proc getRsaHashCfg(alg: JwtAlgorithm): HashCfg =
  let hashAlg = case alg:
    of RS256: SHA256
    of RS384: SHA384
    of RS512: SHA512
    else: raise newException(ValueError,
      "Unsupported RSA signature algorithm '" & $alg & "'")

  result = HASHES[hashAlg]

proc bearRsaSign(
    message: string,
    alg: JwtAlgorithm,
    key: RsaPrivateKeyObj
  ): string =

  let hashCfg = getRsaHashCfg(alg)
  let hashed = hash(message, hashCfg.alg)

  let rsaSignImpl = rsaPkcs1SignGetDefault()
  result = newString((key.bearKey.nBitLen + 7) div 8)

  let errCode = rsaSignImpl(
    cast[ptr cuchar](hashCfg.oid),
    cast[ptr cuchar](unsafeAddr hashed[0]),
    hashed.len,
    unsafeAddr key.bearKey,
    cast[ptr cuchar](addr result[0]))

  if errCode != 1: raise newException(Exception, "RSA signature failed")

proc bearRsaVerify(
    message, signature: string,
    alg: JwtAlgorithm,
    key: RsaPublicKeyObj
  ): bool =

  let hashCfg = getRsaHashCfg(alg)
  let hashed = hash(message, hashCfg.alg)

  let rsaVerifyImpl = rsaPkcs1VrfyGetDefault()
  var recoveredHash = newString(hashCfg.size)

  let errCode = rsaVerifyImpl(
    cast[ptr cuchar](unsafeAddr signature[0]),
    signature.len,
    cast[ptr cuchar](hashCfg.oid),
    hashed.len,
    unsafeAddr key.bearKey,
    cast[ptr cuchar](addr recoveredHash[0]))

  if errCode != 1: return false
  return hashed == recoveredHash

proc rsaSign*(message: string, alg: JwtAlgorithm, key: JWK): string =
  ## Sign a message using the RSA PKCS#1 v1.5 algorithm.
  ##
  ## *key* is expected to be a `JWK <../../jwt/jwk.html#JWK>`_

  if key.keyKind != JwkKeyType.RsaPrivate:
    raise newException(ValueError,
      $alg & " requires an RSA Private key (\"type\"=\"RSA\"), not a " &
      "\"typ\"=\"" & $(key.keyKind) & "\" key.")

  return bearRsaSign(message, alg, toRsaPrivateKey(key))

proc rsaVerify*(message, signature: string; alg: JwtAlgorithm, key: JWK): bool =
  ## Verify the signature for a message using PKCS#1 v1.5 algorithm.
  ##
  ## *key* is expected to be a `JWK <../../jwt/jwk.html#JWK>`_

  if key.keyKind != JwkKeyType.RsaPublic:
    raise newException(ValueError,
      $alg & " requires an RSA Public key (\"type\"=\"RSA\"), not a " &
      "\"typ\"=\"" & $(key.keyKind) & "\" key.")

  return bearRsaVerify(message, signature, alg, toRsaPublicKey(key))
